@@ -1,12 +1,13 @@
-#!/usr/bin/env python3
+#!/usr/bin/env python
 # License: GPL v3 Copyright: 2016, Kovid Goyal <kovid at kovidgoyal.net>
 
 import errno
 import os
 import pwd
 import sys
+from collections.abc import Iterator
 from contextlib import suppress
-from typing import TYPE_CHECKING, Iterable, NamedTuple, Optional, Set
+from typing import TYPE_CHECKING, Any, NamedTuple, Optional
 
 from .types import run_once
 
@@ -22,13 +23,17 @@ class Version(NamedTuple):
 
 appname: str = 'kitty'
 kitty_face = '🐱'
-version: Version = Version(0, 25, 0)
+version: Version = Version(0, 36, 0)
 str_version: str = '.'.join(map(str, version))
 _plat = sys.platform.lower()
 is_macos: bool = 'darwin' in _plat
+is_freebsd: bool = 'freebsd' in _plat
 is_running_from_develop: bool = False
+RC_ENCRYPTION_PROTOCOL_VERSION = '1'
+website_base_url = 'https://sw.kovidgoyal.net/kitty/'
+default_pager_for_help = ('less', '-iRXF')
 if getattr(sys, 'frozen', False):
-    extensions_dir: str = getattr(sys, 'kitty_extensions_dir')
+    extensions_dir: str = getattr(sys, 'kitty_run_data')['extensions_dir']
 
     def get_frozen_base() -> str:
         global is_running_from_develop
@@ -60,10 +65,10 @@ else:
 
 @run_once
 def kitty_exe() -> str:
-    rpath = sys._xoptions.get('bundle_exe_dir')
+    rpath = getattr(sys, 'kitty_run_data').get('bundle_exe_dir')
     if not rpath:
         items = os.environ.get('PATH', '').split(os.pathsep) + [os.path.join(kitty_base_dir, 'kitty', 'launcher')]
-        seen: Set[str] = set()
+        seen: set[str] = set()
         for candidate in filter(None, items):
             if candidate not in seen:
                 seen.add(candidate)
@@ -75,6 +80,11 @@ def kitty_exe() -> str:
     return os.path.join(rpath, 'kitty')
 
 
+@run_once
+def kitten_exe() -> str:
+    return os.path.join(os.path.dirname(kitty_exe()), 'kitten')
+
+
 def _get_config_dir() -> str:
     if 'KITTY_CONFIG_DIRECTORY' in os.environ:
         return os.path.abspath(os.path.expanduser(os.environ['KITTY_CONFIG_DIRECTORY']))
@@ -107,7 +117,7 @@ def _get_config_dir() -> str:
     candidate = os.path.abspath(os.path.expanduser(os.environ.get('XDG_CONFIG_HOME') or '~/.config'))
     ans = os.path.join(candidate, appname)
     try:
-        os.makedirs(ans, exist_ok=True)
+        os.makedirs(os.path.realpath(ans), exist_ok=True)
     except FileExistsError:
         raise SystemExit(f'A file {ans} already exists. It must be a directory, not a file.')
     except PermissionError:
@@ -157,7 +167,7 @@ def runtime_dir() -> str:
     return candidate
 
 
-def wakeup() -> None:
+def wakeup_io_loop() -> None:
     from .fast_data_types import get_boss
     b = get_boss()
     if b is not None:
@@ -168,8 +178,9 @@ terminfo_dir = os.path.join(kitty_base_dir, 'terminfo')
 logo_png_file = os.path.join(kitty_base_dir, 'logo', 'kitty.png')
 beam_cursor_data_file = os.path.join(kitty_base_dir, 'logo', 'beam-cursor.png')
 shell_integration_dir = os.path.join(kitty_base_dir, 'shell-integration')
+fonts_dir = os.path.join(kitty_base_dir, 'fonts')
 try:
-    shell_path = pwd.getpwuid(os.geteuid()).pw_shell or '/bin/sh'
+    shell_path = os.environ.get('SHELL') or pwd.getpwuid(os.geteuid()).pw_shell or '/bin/sh'
 except KeyError:
     with suppress(Exception):
         print('Failed to read login shell via getpwuid() for current user, falling back to /bin/sh', file=sys.stderr)
@@ -178,6 +189,31 @@ except KeyError:
 # https://github.com/ansible/ansible/issues/11536#issuecomment-153030743
 ssh_control_master_template = 'kssh-{kitty_pid}-{ssh_placeholder}'
 
+# See https://specifications.freedesktop.org/icon-naming-spec/latest/ar01s04.html
+# Update the spec in docs/desktop-notifications.rst if you change this.
+standard_icon_names = {
+    'error': ('dialog-error', '☠'),
+    'warning': ('dialog-warning','⚠'),
+    'warn': ('dialog-warning', '⚠'),
+    'info': ('dialog-information', 'ℹ'),
+    'question': ('dialog-question', '❔'),
+
+    'help': ('system-help', '📖'),
+    'file-manager': ('system-file-manager', '🗄'),
+    'system-monitor': ('utilities-system-monitor', '🎛'),
+    'text-editor': ('utilities-text-editor', '📄'),
+}
+
+# See https://github.com/TUNER88/iOSSystemSoundsLibrary for Apple's system
+# sound ids not all of which are available on macOS.
+standard_sound_names = {
+    'error': ('dialog-error', 1),
+    'info': ('dialog-information', 2),
+    'warning': ('dialog-warning', 3),
+    'warn': ('dialog-warning', 3),
+    'question': ('dialog-question', 4),
+}
+
 
 def glfw_path(module: str) -> str:
     prefix = 'kitty.' if getattr(sys, 'frozen', False) else ''
@@ -185,7 +221,7 @@ def glfw_path(module: str) -> str:
 
 
 def detect_if_wayland_ok() -> bool:
-    if 'WAYLAND_DISPLAY' not in os.environ:
+    if 'WAYLAND_DISPLAY' not in os.environ and 'WAYLAND_SOCKET' not in os.environ:
         return False
     if 'KITTY_DISABLE_WAYLAND' in os.environ:
         return False
@@ -217,20 +253,80 @@ def running_in_kitty(set_val: Optional[bool] = None) -> bool:
     return bool(getattr(running_in_kitty, 'ans', False))
 
 
-def list_kitty_resources(package: str = 'kitty') -> Iterable[str]:
-    from importlib.resources import contents
-    return contents(package)
+def list_kitty_resources(package: str = 'kitty') -> Iterator[str]:
+    try:
+        if sys.version_info[:2] < (3, 10):
+            raise ImportError("importlib.resources.files() doesn't work with frozen builds on python 3.9")
+        from importlib.resources import files
+    except ImportError:
+        from importlib.resources import contents
+        return iter(contents(package))
+    else:
+        return (path.name for path in files(package).iterdir())
 
 
 def read_kitty_resource(name: str, package_name: str = 'kitty') -> bytes:
-    from importlib.resources import read_binary
-
-    return read_binary(package_name, name)
+    try:
+        if sys.version_info[:2] < (3, 10):
+            raise ImportError("importlib.resources.files() doesn't work with frozen builds on python 3.9")
+        from importlib.resources import files
+    except ImportError:
+        from importlib.resources import read_binary
+        return read_binary(package_name, name)
+    else:
+        return (files(package_name) / name).read_bytes()
 
 
-def website_url(doc_name: str = '') -> str:
+def website_url(doc_name: str = '', website: str = website_base_url) -> str:
     if doc_name:
-        doc_name = doc_name.rstrip('/')
-        if doc_name:
-            doc_name += '/'
-    return f'https://sw.kovidgoyal.net/kitty/{doc_name}'
+        base, _, frag = doc_name.partition('#')
+        base = base.rstrip('/')
+        if base:
+            base += '/'
+        doc_name = base + (f'#{frag}' if frag else '')
+    return website + doc_name.lstrip('/')
+
+
+handled_signals: set[int] = set()
+
+
+def clear_handled_signals(*a: Any) -> None:
+    if not handled_signals:
+        return
+    import signal
+    if hasattr(signal, 'pthread_sigmask'):
+        signal.pthread_sigmask(signal.SIG_UNBLOCK, handled_signals)
+    for s in handled_signals:
+        signal.signal(s, signal.SIG_DFL)
+
+
+@run_once
+def local_docs() -> str:
+    d = os.path.dirname
+    base = d(d(kitty_exe()))
+    from_source = getattr(sys, 'kitty_run_data').get('from_source')
+    if is_macos and from_source and '/kitty.app/Contents/' in kitty_exe():
+        base = d(d(d(base)))
+    subdir = os.path.join('doc', 'kitty', 'html')
+    linux_ans = os.path.join(base, 'share', subdir)
+    if getattr(sys, 'frozen', False):
+        if is_macos:
+            return os.path.join(d(d(d(extensions_dir))), subdir)
+        return linux_ans
+    if os.path.isdir(linux_ans):
+        return linux_ans
+    if from_source:
+        sq = os.path.join(d(base), 'docs', '_build', 'html')
+        if os.path.isdir(sq):
+            return sq
+    for candidate in ('/usr', '/usr/local', '/opt/homebrew'):
+        q = os.path.join(candidate, 'share', subdir)
+        if os.path.isdir(q):
+            return q
+    return ''
+
+
+@run_once
+def wrapped_kitten_names() -> frozenset[str]:
+    import kitty.fast_data_types as f
+    return frozenset(f.wrapped_kitten_names())
