@@ -1,41 +1,68 @@
-#!/usr/bin/env python3
+#!/usr/bin/env python
 # License: GPL v3 Copyright: 2016, Kovid Goyal <kovid at kovidgoyal.net>
 
 import locale
 import os
 import shutil
 import sys
+from collections.abc import Generator, Sequence
 from contextlib import contextmanager, suppress
-from typing import Dict, Generator, List, Optional, Sequence, Tuple
+from typing import Optional
 
 from .borders import load_borders_program
 from .boss import Boss
-from .child import set_default_env
+from .child import set_default_env, set_LANG_in_default_env
 from .cli import create_opts, parse_args
 from .cli_stub import CLIOptions
 from .conf.utils import BadLine
 from .config import cached_values_for
 from .constants import (
-    appname, beam_cursor_data_file, config_dir, glfw_path, is_macos,
-    is_wayland, kitty_exe, logo_png_file, running_in_kitty
+    appname,
+    beam_cursor_data_file,
+    clear_handled_signals,
+    config_dir,
+    glfw_path,
+    is_macos,
+    is_wayland,
+    kitten_exe,
+    kitty_exe,
+    logo_png_file,
+    running_in_kitty,
+    website_url,
 )
 from .fast_data_types import (
-    GLFW_IBEAM_CURSOR, GLFW_MOD_ALT, GLFW_MOD_SHIFT, create_os_window,
-    free_font_data, glfw_init, glfw_terminate, load_png_data,
-    set_custom_cursor, set_default_window_icon, set_options
+    GLFW_MOD_ALT,
+    GLFW_MOD_SHIFT,
+    SingleKey,
+    create_os_window,
+    free_font_data,
+    glfw_init,
+    glfw_terminate,
+    load_png_data,
+    mask_kitty_signals_process_wide,
+    set_custom_cursor,
+    set_default_window_icon,
+    set_options,
 )
 from .fonts.box_drawing import set_scale
-from .fonts.render import set_font_family
+from .fonts.render import dump_font_debug, set_font_family
 from .options.types import Options
 from .options.utils import DELETE_ENV_VAR
-from .os_window_size import initial_window_size_func
-from .session import get_os_window_sizing_data
-from .types import SingleKey
+from .os_window_size import edge_spacing, initial_window_size_func
+from .session import create_sessions, get_os_window_sizing_data
+from .shaders import CompileError, load_shader_programs
+from .types import LayerShellConfig
 from .utils import (
-    cleanup_ssh_control_masters, detach, expandvars, log_error,
-    single_instance, startup_notification_handler, unix_socket_paths
+    cleanup_ssh_control_masters,
+    detach,
+    expandvars,
+    get_custom_window_icon,
+    log_error,
+    parse_os_window_state,
+    safe_mtime,
+    shlex_split,
+    startup_notification_handler,
 )
-from .window import load_shader_programs
 
 
 def set_custom_ibeam_cursor() -> None:
@@ -48,71 +75,36 @@ def set_custom_ibeam_cursor() -> None:
     rgba_data2, width2, height2 = load_png_data(data)
     images = (rgba_data, width, height), (rgba_data2, width2, height2)
     try:
-        set_custom_cursor(GLFW_IBEAM_CURSOR, images, 4, 8)
+        set_custom_cursor("beam", images, 4, 8)
     except Exception as e:
         log_error(f'Failed to set custom beam cursor with error: {e}')
 
 
-def talk_to_instance(args: CLIOptions) -> None:
-    import json
-    import socket
-    stdin = ''
-    if args.session == '-':
-        stdin = sys.stdin.read()
-    data = {'cmd': 'new_instance', 'args': tuple(sys.argv), 'cmdline_args_for_open': getattr(sys, 'cmdline_args_for_open', []),
-            'startup_id': os.environ.get('DESKTOP_STARTUP_ID'),
-            'cwd': os.getcwd(), 'stdin': stdin}
-    notify_socket = None
-    if args.wait_for_single_instance_window_close:
-        address = f'\0{appname}-os-window-close-notify-{os.getpid()}-{os.geteuid()}'
-        notify_socket = socket.socket(family=socket.AF_UNIX)
-        try:
-            notify_socket.bind(address)
-        except FileNotFoundError:
-            for address in unix_socket_paths(address[1:], ext='.sock'):
-                notify_socket.bind(address)
-                break
-        data['notify_on_os_window_death'] = address
-        notify_socket.listen()
-
-    sdata = json.dumps(data, ensure_ascii=False).encode('utf-8')
-    assert single_instance.socket is not None
-    single_instance.socket.sendall(sdata)
-    with suppress(OSError):
-        single_instance.socket.shutdown(socket.SHUT_RDWR)
-    single_instance.socket.close()
-
-    if args.wait_for_single_instance_window_close:
-        assert notify_socket is not None
-        conn = notify_socket.accept()[0]
-        conn.recv(1)
-        with suppress(OSError):
-            conn.shutdown(socket.SHUT_RDWR)
-        conn.close()
-
-
 def load_all_shaders(semi_transparent: bool = False) -> None:
-    load_shader_programs(semi_transparent)
-    load_borders_program()
+    try:
+        load_shader_programs(semi_transparent)
+        load_borders_program()
+    except CompileError as err:
+        raise SystemExit(err)
 
 
-def init_glfw_module(glfw_module: str, debug_keyboard: bool = False, debug_rendering: bool = False) -> None:
-    if not glfw_init(glfw_path(glfw_module), debug_keyboard, debug_rendering):
+def init_glfw_module(glfw_module: str, debug_keyboard: bool = False, debug_rendering: bool = False, wayland_enable_ime: bool = True) -> None:
+    if not glfw_init(glfw_path(glfw_module), edge_spacing, debug_keyboard, debug_rendering, wayland_enable_ime):
         raise SystemExit('GLFW initialization failed')
 
 
 def init_glfw(opts: Options, debug_keyboard: bool = False, debug_rendering: bool = False) -> str:
     glfw_module = 'cocoa' if is_macos else ('wayland' if is_wayland(opts) else 'x11')
-    init_glfw_module(glfw_module, debug_keyboard, debug_rendering)
+    init_glfw_module(glfw_module, debug_keyboard, debug_rendering, wayland_enable_ime=opts.wayland_enable_ime)
     return glfw_module
 
 
 def get_macos_shortcut_for(
-    func_map: Dict[Tuple[str, ...], List[SingleKey]], defn: str = 'new_os_window', lookup_name: str = ''
+    func_map: dict[tuple[str, ...], list[SingleKey]], defn: str = 'new_os_window', lookup_name: str = ''
 ) -> Optional[SingleKey]:
     # for maximum robustness we should use opts.alias_map to resolve
     # aliases however this requires parsing everything on startup which could
-    # be potentially slow. Lets just hope the user doesnt alias these
+    # be potentially slow. Lets just hope the user doesn't alias these
     # functions.
     ans = None
     candidates = []
@@ -134,23 +126,62 @@ def get_macos_shortcut_for(
     return ans
 
 
-def set_x11_window_icon() -> None:
+def set_macos_app_custom_icon() -> None:
+    custom_icon_mtime, custom_icon_path = get_custom_window_icon()
+    if custom_icon_mtime is not None and custom_icon_path is not None:
+        from .fast_data_types import cocoa_set_app_icon, cocoa_set_dock_icon
+        krd = getattr(sys, 'kitty_run_data')
+        bundle_path = os.path.dirname(os.path.dirname(krd.get('bundle_exe_dir')))
+        icon_sentinel = os.path.join(bundle_path, 'Icon\r')
+        sentinel_mtime = safe_mtime(icon_sentinel)
+        if sentinel_mtime is None or sentinel_mtime < custom_icon_mtime:
+            try:
+                cocoa_set_app_icon(custom_icon_path, bundle_path)
+            except (FileNotFoundError, OSError) as e:
+                log_error(str(e))
+                log_error('Failed to set custom app icon, ignoring')
+        # macOS Dock does not reload icons until it is restarted, so we set
+        # the application icon here. This will revert when kitty quits, but
+        # can't be helped since there appears to be no way to get the dock
+        # to reload short of killing it.
+        cocoa_set_dock_icon(custom_icon_path)
+
+
+def get_icon128_path(base_path: str) -> str:
     # max icon size on X11 64bits is 128x128
-    path, ext = os.path.splitext(logo_png_file)
-    set_default_window_icon(f'{path}-128{ext}')
+    path, ext = os.path.splitext(base_path)
+    return f'{path}-128{ext}'
 
 
-def _run_app(opts: Options, args: CLIOptions, bad_lines: Sequence[BadLine] = ()) -> None:
-    global_shortcuts: Dict[str, SingleKey] = {}
+def set_x11_window_icon() -> None:
+    custom_icon_path = get_custom_window_icon()[1]
+    try:
+        if custom_icon_path is not None:
+            custom_icon128_path = get_icon128_path(custom_icon_path)
+            if safe_mtime(custom_icon128_path) is None:
+                set_default_window_icon(custom_icon_path)
+            else:
+                set_default_window_icon(custom_icon128_path)
+        else:
+            set_default_window_icon(get_icon128_path(logo_png_file))
+    except ValueError as err:
+        log_error(err)
+
+
+def set_cocoa_global_shortcuts(opts: Options) -> dict[str, SingleKey]:
+    global_shortcuts: dict[str, SingleKey] = {}
     if is_macos:
         from collections import defaultdict
         func_map = defaultdict(list)
-        for k, v in opts.keymap.items():
-            parts = tuple(v.split())
-            func_map[parts].append(k)
+        for single_key, v in opts.keyboard_modes[''].keymap.items():
+            kd = v[-1]  # the last definition is the active one
+            if kd.is_suitable_for_global_shortcut:
+                parts = tuple(kd.definition.split())
+                func_map[parts].append(single_key)
 
         for ac in ('new_os_window', 'close_os_window', 'close_tab', 'edit_config_file', 'previous_tab',
-                   'next_tab', 'new_tab', 'new_window', 'close_window', 'toggle_macos_secure_keyboard_entry', 'toggle_fullscreen'):
+                   'next_tab', 'new_tab', 'new_window', 'close_window', 'toggle_macos_secure_keyboard_entry', 'toggle_fullscreen',
+                   'hide_macos_app', 'hide_macos_other_apps', 'minimize_macos_window', 'quit'):
             val = get_macos_shortcut_for(func_map, ac)
             if val is not None:
                 global_shortcuts[ac] = val
@@ -163,21 +194,43 @@ def _run_app(opts: Options, args: CLIOptions, bad_lines: Sequence[BadLine] = ())
         val = get_macos_shortcut_for(func_map, 'load_config_file', lookup_name='reload_config')
         if val is not None:
             global_shortcuts['reload_config'] = val
-    if is_macos and opts.macos_custom_beam_cursor:
-        set_custom_ibeam_cursor()
-    if not is_wayland() and not is_macos:  # no window icons on wayland
-        set_x11_window_icon()
+        val = get_macos_shortcut_for(func_map, f'open_url {website_url()}', lookup_name='open_kitty_website')
+        if val is not None:
+            global_shortcuts['open_kitty_website'] = val
+    return global_shortcuts
+
+
+def _run_app(opts: Options, args: CLIOptions, bad_lines: Sequence[BadLine] = (), talk_fd: int = -1) -> None:
+    if is_macos:
+        global_shortcuts = set_cocoa_global_shortcuts(opts)
+        if opts.macos_custom_beam_cursor:
+            set_custom_ibeam_cursor()
+        set_macos_app_custom_icon()
+    else:
+        global_shortcuts = {}
+        if not is_wayland():  # no window icons on wayland
+            set_x11_window_icon()
+
     with cached_values_for(run_app.cached_values_name) as cached_values:
+        startup_sessions = tuple(create_sessions(opts, args, default_session=opts.startup_session))
+        wincls = (startup_sessions[0].os_window_class if startup_sessions else '') or args.cls or appname
+        window_state = (args.start_as if args.start_as and args.start_as != 'normal' else None) or (
+            getattr(startup_sessions[0], 'os_window_state', None) if startup_sessions else None
+        )
+        wstate = parse_os_window_state(window_state) if window_state is not None else None
         with startup_notification_handler(extra_callback=run_app.first_window_callback) as pre_show_callback:
             window_id = create_os_window(
-                    run_app.initial_window_size_func(get_os_window_sizing_data(opts), cached_values),
+                    run_app.initial_window_size_func(get_os_window_sizing_data(opts, startup_sessions[0] if startup_sessions else None), cached_values),
                     pre_show_callback,
                     args.title or appname, args.name or args.cls or appname,
-                    args.cls or appname, load_all_shaders, disallow_override_title=bool(args.title))
-        boss = Boss(opts, args, cached_values, global_shortcuts)
-        boss.start(window_id)
-        if bad_lines:
-            boss.show_bad_config_lines(bad_lines)
+                    wincls, wstate, load_all_shaders, disallow_override_title=bool(args.title), layer_shell_config=run_app.layer_shell_config)
+        boss = Boss(opts, args, cached_values, global_shortcuts, talk_fd)
+        boss.start(window_id, startup_sessions)
+        if args.debug_font_fallback:
+            dump_font_debug()
+        if bad_lines or boss.misc_config_errors:
+            boss.show_bad_config_lines(bad_lines, boss.misc_config_errors)
+            boss.misc_config_errors = []
         try:
             boss.child_monitor.main_loop()
         finally:
@@ -189,20 +242,21 @@ class AppRunner:
     def __init__(self) -> None:
         self.cached_values_name = 'main'
         self.first_window_callback = lambda window_handle: None
+        self.layer_shell_config: Optional[LayerShellConfig] = None
         self.initial_window_size_func = initial_window_size_func
 
-    def __call__(self, opts: Options, args: CLIOptions, bad_lines: Sequence[BadLine] = ()) -> None:
+    def __call__(self, opts: Options, args: CLIOptions, bad_lines: Sequence[BadLine] = (), talk_fd: int = -1) -> None:
         set_scale(opts.box_drawing_scale)
         set_options(opts, is_wayland(), args.debug_rendering, args.debug_font_fallback)
         try:
-            set_font_family(opts, debug_font_matching=args.debug_font_fallback)
-            _run_app(opts, args, bad_lines)
+            set_font_family(opts, add_builtin_nerd_font=True)
+            _run_app(opts, args, bad_lines, talk_fd)
         finally:
             set_options(None)
             free_font_data()  # must free font data before glfw/freetype/fontconfig/opengl etc are finalized
             if is_macos:
                 from kitty.fast_data_types import (
-                    cocoa_set_notification_activated_callback
+                    cocoa_set_notification_activated_callback,
                 )
                 cocoa_set_notification_activated_callback(None)
 
@@ -215,18 +269,24 @@ def ensure_macos_locale() -> None:
     # https://github.com/kovidgoyal/kitty/issues/90
     from .fast_data_types import cocoa_get_lang, locale_is_valid
     if 'LANG' not in os.environ:
-        lang = cocoa_get_lang()
-        if lang is not None:
-            if not locale_is_valid(lang):
-                if lang.startswith('en_'):
-                    lang = 'en_US'
-                else:
-                    log_error(f'Could not set LANG Cocoa returns language as: {lang}')
-            os.environ['LANG'] = f'{lang}.UTF-8'
+        lang_code, country_code, identifier = cocoa_get_lang()
+        lang = 'en_US'
+        if identifier and locale_is_valid(identifier):
+            lang = identifier
+        elif lang_code and country_code and locale_is_valid(f'{lang_code}_{country_code}'):
+            lang = f'{lang_code}_{country_code}'
+        elif lang_code:
+            if lang_code != 'en':
+                with suppress(OSError):
+                    found = sorted(x for x in os.listdir('/usr/share/locale') if x.startswith(f'{lang_code}_'))
+                    if found:
+                        lang = found[0].partition('.')[0]
+        os.environ['LANG'] = f'{lang}.UTF-8'
+        set_LANG_in_default_env(os.environ['LANG'])
 
 
 @contextmanager
-def setup_profiling(args: CLIOptions) -> Generator[None, None, None]:
+def setup_profiling() -> Generator[None, None, None]:
     try:
         from .fast_data_types import start_profiler, stop_profiler
         do_profile = True
@@ -244,29 +304,30 @@ def setup_profiling(args: CLIOptions) -> Generator[None, None, None]:
         with open(cg, 'wb') as f:
             subprocess.call(['pprof', '--callgrind', exe, '/tmp/kitty-profile.log'], stdout=f)
         try:
-            subprocess.Popen(['kcachegrind', cg])
+            subprocess.Popen(['kcachegrind', cg], preexec_fn=clear_handled_signals)
         except FileNotFoundError:
             subprocess.call(['pprof', '--text', exe, '/tmp/kitty-profile.log'])
             print('To view the graphical call data, use: kcachegrind', cg)
 
 
-def macos_cmdline(argv_args: List[str]) -> List[str]:
+def macos_cmdline(argv_args: list[str]) -> list[str]:
     try:
         with open(os.path.join(config_dir, 'macos-launch-services-cmdline')) as f:
             raw = f.read()
     except FileNotFoundError:
         return argv_args
-    import shlex
     raw = raw.strip()
-    ans = shlex.split(raw)
+    ans = list(shlex_split(raw))
     if ans and ans[0] == 'kitty':
         del ans[0]
-    return ans
+    return ans + argv_args
 
 
 def expand_listen_on(listen_on: str, from_config_file: bool) -> str:
+    if from_config_file and listen_on == 'none':
+        return ''
     listen_on = expandvars(listen_on)
-    if '{kitty_pid}' not in listen_on and from_config_file:
+    if '{kitty_pid}' not in listen_on and from_config_file and listen_on.startswith('unix:'):
         listen_on += '-{kitty_pid}'
     listen_on = listen_on.replace('{kitty_pid}', str(os.getpid()))
     if listen_on.startswith('unix:'):
@@ -277,6 +338,9 @@ def expand_listen_on(listen_on: str, from_config_file: bool) -> str:
             elif not os.path.isabs(path):
                 import tempfile
                 listen_on = f'unix:{os.path.join(tempfile.gettempdir(), path)}'
+    elif listen_on.startswith('tcp:') or listen_on.startswith('tcp6:'):
+        if from_config_file:  # use a random port
+            listen_on = ':'.join(listen_on.split(':', 2)[:2]) + ':0'
     return listen_on
 
 
@@ -298,11 +362,12 @@ def prepend_if_not_present(path: str, paths_serialized: str) -> str:
 
 def ensure_kitty_in_path() -> None:
     # Ensure the correct kitty is in PATH
-    rpath = sys._xoptions.get('bundle_exe_dir')
+    krd = getattr(sys, 'kitty_run_data')
+    rpath = krd.get('bundle_exe_dir')
     if not rpath:
         return
     if rpath:
-        modify_path = is_macos or getattr(sys, 'frozen', False) or sys._xoptions.get('kitty_from_source') == '1'
+        modify_path = is_macos or getattr(sys, 'frozen', False) or krd.get('from_source')
         existing = shutil.which('kitty')
         if modify_path or not existing:
             env_path = os.environ.get('PATH', '')
@@ -311,16 +376,41 @@ def ensure_kitty_in_path() -> None:
                 os.environ['PATH'] = prepend_if_not_present(rpath, env_path)
 
 
+def ensure_kitten_in_path() -> None:
+    correct_kitten = kitten_exe()
+    existing = shutil.which('kitten')
+    if existing and safe_samefile(existing, correct_kitten):
+        return
+    env_path = os.environ.get('PATH', '')
+    os.environ['PATH'] = prepend_if_not_present(os.path.dirname(correct_kitten), env_path)
+
+
+def setup_manpath(env: dict[str, str]) -> None:
+    # Ensure kitty manpages are available in frozen builds
+    if not getattr(sys, 'frozen', False):
+        return
+    from .constants import local_docs
+    mp = os.environ.get('MANPATH', env.get('MANPATH', ''))
+    d = os.path.dirname
+    kitty_man = os.path.join(d(d(d(local_docs()))), 'man')
+    if not mp:
+        env['MANPATH'] = f'{kitty_man}:'
+    elif mp.startswith(':'):
+        env['MANPATH'] = f':{kitty_man}:{mp}'
+    else:
+        env['MANPATH'] = f'{kitty_man}:{mp}'
+
+
 def setup_environment(opts: Options, cli_opts: CLIOptions) -> None:
     from_config_file = False
-    if not cli_opts.listen_on and opts.listen_on.startswith('unix:'):
+    if not cli_opts.listen_on:
         cli_opts.listen_on = opts.listen_on
         from_config_file = True
-    if cli_opts.listen_on and opts.allow_remote_control != 'n':
+    if cli_opts.listen_on:
         cli_opts.listen_on = expand_listen_on(cli_opts.listen_on, from_config_file)
-        os.environ['KITTY_LISTEN_ON'] = cli_opts.listen_on
     env = opts.env.copy()
     ensure_kitty_in_path()
+    ensure_kitten_in_path()
     kitty_path = shutil.which('kitty')
     if kitty_path:
         child_path = env.get('PATH')
@@ -328,6 +418,7 @@ def setup_environment(opts: Options, cli_opts: CLIOptions) -> None:
         # the other values mean the user doesn't want a PATH
         if child_path not in ('', DELETE_ENV_VAR) and child_path is not None:
             env['PATH'] = prepend_if_not_present(os.path.dirname(kitty_path), env['PATH'])
+    setup_manpath(env)
     set_default_env(env)
 
 
@@ -345,22 +436,17 @@ def set_locale() -> None:
             except Exception:
                 log_error('Failed to set locale with no LANG')
             os.environ['LANG'] = old_lang
+            set_LANG_in_default_env(old_lang)
 
 
 def _main() -> None:
     running_in_kitty(True)
-    with suppress(AttributeError):  # python compiled without threading
-        sys.setswitchinterval(1000.0)  # we have only a single python thread
-
-    try:
-        set_locale()
-    except Exception:
-        log_error('Failed to set locale, ignoring')
 
     args = sys.argv[1:]
     if is_macos and os.environ.pop('KITTY_LAUNCHED_BY_LAUNCH_SERVICES', None) == '1':
         os.chdir(os.path.expanduser('~'))
         args = macos_cmdline(args)
+        getattr(sys, 'kitty_run_data')['launched_by_launch_services'] = True
     try:
         cwd_ok = os.path.isdir(os.getcwd())
     except Exception:
@@ -385,29 +471,52 @@ def _main() -> None:
     if cli_opts.detach:
         if cli_opts.session == '-':
             from .session import PreReadSession
-            cli_opts.session = PreReadSession(sys.stdin.read())
+            cli_opts.session = PreReadSession(sys.stdin.read(), os.environ)
         detach()
     if cli_opts.replay_commands:
         from kitty.client import main as client_main
         client_main(cli_opts.replay_commands)
         return
+    talk_fd = -1
     if cli_opts.single_instance:
-        is_first = single_instance(cli_opts.instance_group)
-        if not is_first:
-            talk_to_instance(cli_opts)
-            return
-    bad_lines: List[BadLine] = []
+        si_data = os.environ.pop('KITTY_SI_DATA', '')
+        if si_data:
+            import atexit
+            fdnum, sep, socket_path = si_data.partition(':')
+            talk_fd = int(fdnum)
+            def cleanup_si() -> None:
+                with suppress(OSError):
+                    os.close(talk_fd)
+                with suppress(OSError):
+                    if sep and socket_path:
+                        os.unlink(socket_path)
+            atexit.register(cleanup_si)
+    bad_lines: list[BadLine] = []
     opts = create_opts(cli_opts, accumulate_bad_lines=bad_lines)
-    init_glfw(opts, cli_opts.debug_keyboard, cli_opts.debug_rendering)
     setup_environment(opts, cli_opts)
+
+    # set_locale on macOS uses cocoa APIs when LANG is not set, so we have to
+    # call it after the fork
+    try:
+        set_locale()
+    except Exception:
+        log_error('Failed to set locale, ignoring')
+    with suppress(AttributeError):  # python compiled without threading
+        sys.setswitchinterval(1000.0)  # we have only a single python thread
+
     if cli_opts.watcher:
         from .window import global_watchers
         global_watchers.set_extra(cli_opts.watcher)
         log_error('The --watcher command line option has been deprecated in favor of using the watcher option in kitty.conf')
+    # mask the signals now as on some platforms the display backend starts
+    # threads. These threads must not handle the masked signals, to ensure
+    # kitty can handle them. See https://github.com/kovidgoyal/kitty/issues/4636
+    mask_kitty_signals_process_wide()
+    init_glfw(opts, cli_opts.debug_keyboard, cli_opts.debug_rendering)
     try:
-        with setup_profiling(cli_opts):
+        with setup_profiling():
             # Avoid needing to launch threads to reap zombies
-            run_app(opts, cli_opts, bad_lines)
+            run_app(opts, cli_opts, bad_lines, talk_fd)
     finally:
         glfw_terminate()
         cleanup_ssh_control_masters()
